# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Module for Jupyter Widget that displays the JSME editor.
"""
import uuid
import ipywidgets

from collections import namedtuple
from typing import TYPE_CHECKING
from notebook.nbextensions import check_nbextension
from varname import varname, VarnameRetrievingError
from IPython.core.display import Javascript, display, HTML

from rdkit.Chem import AllChem as Chem

if TYPE_CHECKING:
    from rdkit.Chem import Mol

# Named tuple for storing serialized JSME widget value in several
# popular formats
JsmeValue = namedtuple("JsmeValue", ["jme", "molblock", "smiles"])

# Rel file path for Javascript source
JS_SOURCE = "jsme/jsme.nocache.js"

# Base URL for widget source javascript
BASE_URL = "/nbextensions/jupyter_jsme"

if check_nbextension("jupyter_jsme") is False:
    # Fallback option in case user didn't install Jupyter extension
    BASE_URL = "https://peter-ertl.com/jsme/JSME_2020-06-11"

_HTML_STR_FORMAT = '''
<script type="text/javascript" src="{base_url}/{js_source}"></script>
<script type="text/javascript">
function jsmeOnLoad() {{
    jsmeApp_{uid} = new JSApplet.JSME("JSApp_{uid}", \
        "{width}px", "{height}px", {{
        "options" : "{options}",
        "{data_type}": "{data}"
    }});
}}

function saveAs_{uid}(var_name) {{
    var molblock = jsmeApp_{uid}.molFile();
    var smiles = jsmeApp_{uid}.smiles();
    var jme = jsmeApp_{uid}.jmeFile();
    var command = var_name + '.set_value(jme="' + jme + '", smiles="' + \
        smiles + '", molblock="""' + molblock + '""")';
    console.log(command);
    IPython.notebook.kernel.execute(command);
}}

</script>
<div id="JSApp_{uid}"></div>
'''


class JsmeWidgetButton(ipywidgets.Button):
    def __init__(self, uid: str, *args, **kwargs):
        """Create "save" button for JsmeWidget

        :param uid: Unique ID of widget
        :type uid: str
        """
        self.uid = uid
        return super(JsmeWidgetButton, self).__init__(
            description="Save", *args, **kwargs)


class JsmeWidget:
    """Jupyter widget for displaying JSME molecular editor.
    Allows to either build molecule from scratch and save the JME, SMILES and
    MolBlock values, or set either of the aforementioned values manually.
    """
    n = 0

    def __init__(
        self,
        width=400,
        height=350,
        jme="",
        smiles="",
        molblock="",
        options="query,hydrogens",
        parent_varname: str = ""
    ):
        """Create JsmeWidget instance

        :param width: Widget width in pixels, defaults to 400
        :type width: int, optional
        :param height: Widget height in pixels, defaults to 350
        :type height: int, optional
        :param jme: JME value string, defaults to ""
        :type jme: str, optional
        :param smiles: SMILES value string, defaults to ""
        :type smiles: str, optional
        :param molblock: MolBlock value string, defaults to ""
        :type molblock: str, optional
        :param options: Options to pass to widget, defaults to "query,
        hydrogens"
        :type options: str, optional
        """
        try:
            self.name = f"{parent_varname}.{varname()}"
        except VarnameRetrievingError:
            self.name = "_"

        self.width = width
        self.height = height
        self.value = JsmeValue(jme=jme, molblock=molblock, smiles=smiles)
        self.options = options
        self._uids = []
        self._updated = False

    def set_value(self, jme: str, smiles: str, molblock: str):
        """Set JSME value (JME, SMILES and MolBlock)

        :param jme: JME value string
        :type jme: str
        :param smiles: SMILES value string
        :type smiles: str
        :param molblock: MolBlock value string
        :type molblock: str
        """
        self._updated = True
        self.value = JsmeValue(jme=jme, molblock=molblock, smiles=smiles)

    @property
    def was_updated(self):
        return self._updated

    def reset_updated(self):
        self._updated = False

    def _gen_uid(self):
        """Generate unique identifier for javascript applet"""
        uid = str(uuid.uuid1()).replace("-", "")
        # Keep track of all UIDs
        self._uids.append(uid)
        return uid

    def html_str(self, uid: str) -> str:
        """Returns an HTML string that contains the widget.

        :param uid: Unique identifier of widget
        :type uid: str
        :return: HTML string for displaying widget
        :rtype: str
        """
        JsmeWidget.n += 1
        data_type = "jme" if self.value.jme else "smiles"
        data = self.value.jme if self.value.jme else self.value.smiles

        return _HTML_STR_FORMAT.format(
            base_url=BASE_URL,
            js_source=JS_SOURCE,
            var_name=self.name,
            uid=uid,
            width=self.width,
            height=self.height,
            options=self.options,
            data_type=data_type,
            data=data
        )

    def _ipython_display_(self):
        """Display the widget"""
        uid = self._gen_uid()
        jsme = HTML(self.html_str(uid=uid))
        button = JsmeWidgetButton(uid=uid)

        def _save(button):
            self.save(button.uid)

        button.on_click(_save)
        display(jsme, button)

    def save(self, uid):
        """Save the value of the molecular diagram drawn in the
        widget to the Python namespace
        """
        display(Javascript(f"saveAs_{uid}('{self.name}')"))

    def to_mol(self, add_hs: bool = False, num_confs: int = 10) -> "Mol":
        """Convert widget value to RDKit molecule.
        If Hydrogen atoms are added, calculate the optimal conformer to get
        Hydrogen atom coordinates.

        :param add_hs: Add Hydrogen atoms
        :type add_hs: bool
        :param num_confs: Number of conformers to generate
        :type num_confs: int
        :return: RDKit molecule object
        :rtype: Mol
        """
        if self.value.smiles != "":
            mol = Chem.MolFromSmiles(self.value.smiles)
        elif self.value.molblock != "":
            mol = Chem.MolFromMolBlock(self.value.molblock)
        else:
            raise ValueError("Cannot create Mol object: JSME value is empty")

        if add_hs is True and mol is not None:
            mol = Chem.AddHs(mol)
            # Calculate conformers after adding hydrogens
            Chem.EmbedMultipleConfs(mol, numConfs=num_confs)

        if mol is None:
            raise ValueError(f"Cannot convert JSME widget value to Mol object:\
                \n{self.value}\
                \nPlease try clicking 'Save' button on widget or remove \
                Hydrogen atoms from SMILES string and try again.")
        return mol



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import pytest
from qdk.chemistry.geometry import Element, Geometry


def test_Element():
    c1 = Element(
        name="C",
        x=0.0,
        y=0.0,
        z=0.0
    )

    c2 = Element(
        name="C",
        x=1.0,
        y=0.0,
        z=0.0
    )

    g = Geometry([c1, c2])
    assert g[0] == c1
    assert g[1] == c2


@pytest.fixture
def xyz():
    return """3
water
O 0.002 0.398 0.0
H 0.762 -0.203 0.0
H -0.764 -0.195 0.0"""


def test_geometry_from_mol(h2o):
    g = Geometry.from_mol(h2o)
    assert g.charge == 0
    assert len(g) == 3
    assert [el.name for el in g] == ["O", "H", "H"]


def test_geometry_to_from_xyz(geometry, xyz):
    assert geometry.to_xyz("water") == xyz
    gprime = Geometry.from_xyz(xyz)
    assert geometry == gprime



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

##
# Module for loading and encoding Broombridge data
##
import logging
from qsharp.chemistry import load_broombridge, load_input_state, encode
from typing import List, Tuple

NumQubits = int
HamiltonianTermList = Tuple[List[Tuple[List[int], List[float]]]]
InputStateTerms = Tuple[int, List[Tuple[Tuple[float, float], List[int]]]]
EnergyOffset = float
JWEncodedData = Tuple[
    NumQubits,
    HamiltonianTermList,
    InputStateTerms,
    EnergyOffset
]

_log = logging.getLogger(__name__)


def load_and_encode(
    file_name: str,
    problem_description_index: int = 0,
    initial_state_label: str = None
) -> JWEncodedData:
    """Wrapper function for loading and encoding Broombridge file into
    JWEncodedData-compatible format.

    :param file_name: Broombridge file name
    :type file_name: str
    :param problem_description_index: Index of problem description to use,
        defaults to 0
    :type problem_description_index: int, optional
    :param initial_state_label: Label of initial state to use, defaults to
        first available label
    :type initial_state_label: str, optional
    """
    broombridge_data = load_broombridge(file_name)
    problem = broombridge_data.problem_description[problem_description_index]

    if initial_state_label is None:
        # Pick first in list
        initial_state_label = problem.initial_state_suggestions[0].get("Label")
        _log.info(f"Using initial state label: {initial_state_label}")

    input_state = load_input_state(file_name, initial_state_label)
    ferm_hamiltonian = problem.load_fermion_hamiltonian()
    (
        num_qubits,
        hamiltonian_term_list,
        input_state_terms,
        energy_offset
    ) = encode(ferm_hamiltonian, input_state)

    return (
        num_qubits,
        hamiltonian_term_list,
        input_state_terms,
        energy_offset
    )



import os
import numpy as np
import pytest
from rdkit import Chem, rdBase
from rdkit.Chem import AllChem, rdmolops

from qdk.chemistry._xyz2mol.util import (
    get_atoms, 
    get_mol, 
    generate_structure_from_smiles, 
    get_proto_mol,
    read_xyz_file
)

from qdk.chemistry._xyz2mol.xyz2mol import xyz2mol
from qdk.chemistry._xyz2mol.ac import AC2mol

__TEST_SMILES__ = [
    'C[C-](c1ccccc1)C',
    'C[C-](C)c1ccccc1',
    'C=C([O-])CC',
    'C=C([NH3+])CC',
    'CC(=O)[O-]',
    'C[N+](=O)[O-]',
    'CS(CC)(=O)=O',
    'CS([O-])(=O)=O',
    'C=C(C)CC',
    'CC(C)CC',
    'C=C(N)CC',
    'C=C(C)C=C',
    'C#CC=C',
    'c1ccccc1',
    'c1ccccc1c1ccccc1',
    '[NH3+]CS([O-])(=O)=O',
    'CC(NC)=O',
    '[O-]c1ccccc1',
    'O=C(C=C1)C=CC1=CCC([O-])=O',
    'C#CC#C',
    'Cc1ccc(cc1)C1C=CC2C(C=CC2(C#N)C#N)=CC=1',
    '[C+](C)(C)CC[C-](C)(C)',
    'O=C(C=C1)C=CC1=CCC([O-])=O',
    '[O-]c1ccccc1',
    'Cc1ccc(cc1)C1C=CC2C(C=CC2(C#N)C#N)=CC=1',
    'CC1C=CC2C(C=CC2(C)C)=CC=1',
    'CC1=CC=C(C=CC2)C2C=C1',
    'CC1=CC=C(C2=CC=CC=C2)C=C1',
    'C1(CC2=CC=CC=C2)=CC=CC=C1',
    '[O-]c1ccccc1[O-]',
    'C[N+](=O)[O-]',
    'N#CC(C#N)=CC=C1C=CC=CC(=C1)c1ccc(cc1)[N+](=O)[O-]',
    'CNC([O-])=C([NH+]=C/CC(O)=O)C'
]

__TEST_FILES__ = [
    ("ethane.xyz", 0, "CC"),
    # ("examples/acetate.xyz", -1, "CC(=O)[O-]"),
    # ("examples/chiral_stereo_test.xyz", 0, "C/C=C/[C@@H](C)F"),
    # ("examples/propylbenzene.xyz", -1, "C[C-](C)c1ccccc1"),
]


@pytest.mark.parametrize("smiles", __TEST_SMILES__)
def test_smiles_from_adjacent_matrix(smiles):

    charged_fragments = True
    quick = True

    # Cut apart the smiles
    mol = get_mol(smiles)
    atoms = get_atoms(mol)
    charge = Chem.GetFormalCharge(mol)
    adjacent_matrix = Chem.GetAdjacencyMatrix(mol)

    #
    mol = Chem.RemoveHs(mol)
    canonical_smiles = Chem.MolToSmiles(mol)

    # Define new molecule template from atoms
    new_mol = get_proto_mol(atoms)

        # reconstruct the molecule from adjacent matrix, atoms and total charge
    new_mols = AC2mol(new_mol, adjacent_matrix, atoms, charge, charged_fragments, quick)
    
    new_mol_smiles_list = []
    for new_mol in new_mols:
        new_mol = Chem.RemoveHs(new_mol)
        new_mol_smiles = Chem.MolToSmiles(new_mol)

        new_mol_smiles_list.append(new_mol_smiles)

    assert canonical_smiles in new_mol_smiles_list

    return

@pytest.mark.parametrize("smiles", __TEST_SMILES__)
def test_smiles_from_coord_vdw(smiles):

    # The answer
    mol = Chem.MolFromSmiles(smiles)
    charge = Chem.GetFormalCharge(mol)
    canonical_smiles = Chem.MolToSmiles(mol, isomericSmiles=False)

    # generate forcefield coordinates
    atoms, coordinates = generate_structure_from_smiles(smiles)

    # Generate molobj from atoms, charge and coordinates
    mols = xyz2mol(atoms, coordinates, charge=charge)

    smiles_list = []
    for mol in mols:
    # For this test, remove chira. clean and canonical
        Chem.Kekulize(mol)
        mol = Chem.RemoveHs(mol)
        Chem.RemoveStereochemistry(mol)
        smiles = Chem.MolToSmiles(mol, isomericSmiles=False)

        # Please look away. A small hack that removes the explicit hydrogens
        mol = Chem.MolFromSmiles(smiles)
        smiles = Chem.MolToSmiles(mol)
        smiles_list.append(smiles)

    assert canonical_smiles in smiles_list

    return


@pytest.mark.parametrize("smiles", __TEST_SMILES__)
def test_smiles_from_coord_huckel(smiles):

    # The answer
    mol = Chem.MolFromSmiles(smiles)
    charge = Chem.GetFormalCharge(mol)
    canonical_smiles = Chem.MolToSmiles(mol, isomericSmiles=False)

    # generate forcefield coordinates
    atoms, coordinates = generate_structure_from_smiles(smiles)

    # Generate molobj from atoms, charge and coordinates
    mols = xyz2mol(atoms, coordinates, charge=charge, use_huckel=True)

    smiles_list = []
    for mol in mols:
        # For this test, remove chira. clean and canonical
        Chem.Kekulize(mol)
        mol = Chem.RemoveHs(mol)
        Chem.RemoveStereochemistry(mol)
        smiles = Chem.MolToSmiles(mol, isomericSmiles=False)

        # Please look away. A small hack that removes the explicit hydrogens
        mol = Chem.MolFromSmiles(smiles)
        smiles = Chem.MolToSmiles(mol)
        smiles_list.append(smiles)

    assert canonical_smiles in smiles_list

    return


@pytest.mark.parametrize("filename, charge, answer", __TEST_FILES__)
def test_smiles_from_xyz_files(filename, charge, answer):

    charged_fragments = True
    quick = True

    atoms, charge_read, coordinates = read_xyz_file(os.path.join(os.path.split(__file__)[0], filename))
    mols = xyz2mol(atoms, coordinates, charge=charge)

    smiles_list = []
    for mol in mols:
        mol = Chem.RemoveHs(mol)

        smiles = Chem.MolToSmiles(mol)
        smiles_list.append(smiles)

    print(answer, smiles_list)

    assert answer in smiles_list

    return



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Module for generating NWChem input deck format.
"""

import logging
import re
from typing import Dict, Union, TYPE_CHECKING

from qdk.chemistry.geometry import Geometry
from qdk.chemistry.solvers.util import formatted_geometry_str

if TYPE_CHECKING:
    from rdkit.Chem.AllChem import Mol

_log = logging.getLogger(__name__)

__all__ = [
    "parse_nwchem_output",
    "create_input_deck"
]

_GEOMETRY_UNITS = ["au", "angstrom"]

# Template for generating an NWChem input deck
NW_CHEM_TEMPLATE = """
start {name}

echo
{memory}

geometry units {geometry_units}
symmetry c1
{geometry}
end

basis
* library {basis}
end

{charge}

scf
thresh {scf_thresh:.1e}
tol2e {scf_tol2e}
{spin}
{rhf}
maxiter 200
{nopen}
end

tce
{method}
2eorb
2emet 13
tilesize 20
{nroots}
thresh {tce_thresh:.1e}
end

set tce:print_integrals T
set tce:qorb {num_orb}
set tce:qela {num_el_a}
set tce:qelb {num_el_b}

task tce {driver}
"""


def create_input_deck(
    mol: "Mol",
    mol_name: str,
    num_active_orbitals: int,
    geometry: Union[str, Geometry] = None,
    memory: str = "memory stack 1000 mb heap 100 mb global 1000 mb noverify",
    geometry_units: str = "au",
    basis: str = "sto-3g",
    charge: int = 0,
    scf_thresh: float = 1.0e-8,
    scf_tol2e: float = 1.0e-9,
    rhf: str = "rhf",
    spin: str = "singlet",
    nopen: int = None,
    method: str = "ccsd",
    num_tce_root: int = None,
    tce_thresh: float = 1.0e-8,
    driver: str = "energy",
    num_active_el: int = None,
) -> str:
    """Generate an NWChem input deck

    :param mol: RDKit Molecule object to use for calculating number of
        electrons if unspecified, defaults to None
    :type mol: Mol, optional
    :param mol_name: Molecule name
    :type mol_name: str
    :param geometry: Molecule geometry in the following format (each atom on
        a new line) : [Atom] [X] [Y] [Z]
    :type geometry: str
    :param num_active_orbitals: Number of orbitals in molecular ground state
        to set active space
    :type num_active_orbitals: int
    :param memory: Memory specification, defaults to "memory stack 1000 mb
        heap 100 mb global 1000 mb noverify"
    :type memory: str, optional
    :param geometry_units: Units used for geometry ("au"|"angstrom"), defaults
        to "au"
    :type geometry_units: str, optional
    :param basis: Basis to use for atoms, defaults to "sto-3g"
    :type basis: str, optional
    :param charge: Molecule charge, defaults to 0
    :type charge: int, optional
    :param scf_thresh: Threshold for SCF solver with one-decimal precision,
        defaults to 1.0e-10
    :type scf_thresh: float, optional
    :param scf_tol2e: 2-electron tolerance for SCF solver, defaults to 1.0e-10
    :type scf_tol2e: float, optional
    :param rhf: Restricted (Open-shell) Hartree Fock method. Either "rhf" or
        "rohf", defaults to "rhf"
    :type rhf: str, optional
    :param spin: Spin property, defaults to "singlet"
    :type spin: str, optional
    :param nopen: Number of singly occupied electron orbitals, defaults to None
    :type nopen: int, optional
    :param method: Calculation method, either "HF", "DFT" or "ccsd", defaults
        to "ccsd"
    :type method: str, optional
    :param num_tce_root: Number of excited states, defaults to 5
    :type num_tce_root: int, optional
    :param tce_thresh: Threshold for TCE solver with one-decimal precision,
        defaults to 1.0e-6
    :type tce_thresh: float, optional
    :param driver: Driver method, defaults to "energy"
    :type driver: str, optional
    :param num_active_el: Number of active electrons in molecule. This value
        is calculated based on atomic numbers if mol is provided. Defaults to
        None, defaults to None
    :type num_active_el: int, optional
    :raises ValueError: If neither Mol or num_active_el|geometry are specified.
    :return: NWChem input deck formatted string
    :rtype: str
    """
    assert geometry_units in _GEOMETRY_UNITS, f"Unknown geometry unit: \
        {geometry_units}"

    num_active_el = num_active_el or num_active_orbitals
    geometry = formatted_geometry_str(mol=mol, geometry=geometry)
    nopen_str = f"nopen {nopen}" if nopen is not None else ""
    nroots = f"nroots {num_tce_root}" if num_tce_root else ""
    charge_text = f"charge {charge}" if charge else ""

    nw_chem = NW_CHEM_TEMPLATE.format(
        name=f"{mol_name}",
        memory=memory,
        geometry_units=geometry_units,
        geometry=geometry,
        basis=basis,
        charge=charge_text,
        scf_thresh=scf_thresh,
        scf_tol2e=scf_tol2e,
        rhf=rhf,
        spin=spin,
        nopen=nopen_str,
        method=method,
        nroots=nroots,
        tce_thresh=tce_thresh,
        num_orb=num_active_orbitals,
        num_el_a=num_active_el//2,
        num_el_b=num_active_el//2,
        driver=driver
    )

    return nw_chem


def parse_nwchem_output(
    nw_input_file_path: str,
    nw_output_file_path: str
) -> Dict[str, Union[int, float]]:
    """Parse NWChem output

    :param nw_input_file_path: NW input filepath
    :type nw_input_file_path: str
    :param nw_output_file_path: NW output filepath
    :type nw_output_file_path: str
    :return: n_atom, n_orbital, scf_energy, ccsd_corr_energy
    """
    with open(nw_output_file_path) as f:
        logfile = f.read()

    with open(nw_input_file_path) as f:
        nwfile = f.read()

    pattern_atom = "atoms           =\\s*([0-9]*)"
    pattern_orbital = "Number of AO functions :\\s*([0-9]*)"
    pattern_scf = "Total SCF energy =\\s*([+-]?[0-9]*[.][0-9]+)"

    n_atom = int(re.findall(pattern_atom, logfile)[0])
    n_orbital = int(re.findall(pattern_orbital, logfile)[0])
    scf_energy = re.findall(pattern_scf, logfile)

    pattern_ccsd = "correlation energy / hartree =\\s*([+-]?[0-9]*[.][0-9]+)"
    ccsd_corr_energy = re.findall(pattern_ccsd, logfile)

    geometry_snapshot = []

    if "optimize" in nwfile:
        # The following geometry snapshots doesn't
        # include the initial one, but does include final one.
        cd = "[+-]?[0-9]*[.][0-9]+"
        pattern_geom = f"([0-9]+)\\s*([A-Z]+)\\s*({cd})\\s*({cd})\\s*({cd})\
        \\s*({cd})\\s*({cd})\\s*({cd})"
        geomsnapshot = re.findall(pattern_geom, logfile)

        for i in range(0, len(geomsnapshot)):
            if (i % n_atom == 0):
                geometry_snapshot.append(geomsnapshot[i][0:5])
                _log.info("Geometry snapshot %d" % (i/n_atom))
            _log.info(geomsnapshot[i][0:5])

    _log.info("Number of Atoms: ", n_atom)
    _log.info("Number of Orbitals: ", n_orbital)

    scf_energy = float(scf_energy[-1]) if scf_energy else None
    _log.info("SCF Energy: ", scf_energy)
    ccsd_corr_energy = float(ccsd_corr_energy[-1]) if ccsd_corr_energy else \
        None
    _log.info("Correlation Energy: ", ccsd_corr_energy)

    return {
        "number of atoms": n_atom,
        "number of orbitals": n_orbital,
        "SCF energy": scf_energy,
        "CCSD correlation energy": ccsd_corr_energy,
        "geometry snapshot": geometry_snapshot
    }



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

##
# Module for the Molecule class for defining and visualizing molecules
##
import enum
import logging
import os
import re

from collections import namedtuple
from IPython.display import display
from typing import List, Dict

try:
    from rdkit.Chem import AllChem as Chem
except ImportError as e:
    raise ImportError("Missing dependency: rdkit. \
Install with Conda: \n\n\tconda install -c conda-forge rdkit")

from qdk.chemistry.widgets.jsmol_widget import JsmolWidget
from qdk.chemistry.widgets.jsme_widget import JsmeWidget
from qdk.chemistry.geometry import Geometry
from qdk.chemistry.solvers import nwchem, openmolcas, psi4
from qdk.chemistry._xyz2mol import xyz2mol, read_xyz_file
from qdk.chemistry.solvers.util import num_electrons

import basis_set_exchange as bse

_log = logging.getLogger(__name__)

DEFAULT_BASE_PATH = os.environ.get("QDKCHEM_OUTPUT_PATH", ".")
SolverSpec = namedtuple("SolverSpec", ["name", "module", "extension"])


class Solver(enum.Enum):
    nwchem = SolverSpec(
        name="NWChem",
        module=nwchem,
        extension=".nw"
    )
    openmolcas = SolverSpec(
        name="OpenMolcas",
        module=openmolcas,
        extension=".inp"
    )
    psi4 = SolverSpec(
        name="Psi4",
        module=psi4,
        extension=".in"
    )


class Molecule(object):
    """
    Molecule object for visualization and geometry generation
    """
    def __init__(self, mol: Chem.Mol, num_confs: int = 10, xyz: str = None):
        self.mol = mol
        self.design_widget = None
        self._xyz = xyz

        if xyz:
            self.widget = JsmolWidget.from_str(data=xyz)
        elif mol:
            self.widget = JsmolWidget.from_mol(mol=mol, num_confs=num_confs)
        else:
            self.widget = None

    @classmethod
    def from_smiles(
        cls,
        smiles: str,
        add_hs: bool = True,
        num_confs: int = 10
    ):
        mol = Chem.MolFromSmiles(smiles)
        if add_hs:
            mol = Chem.AddHs(mol)
            # Calculate conformers after adding hydrogens
            Chem.EmbedMultipleConfs(mol, numConfs=num_confs)

        return cls(mol=mol)

    @classmethod
    def from_xyz(cls, xyz_file: str, add_hs: bool = True):
        atoms, charge_read, coordinates = read_xyz_file(xyz_file)
        mols = xyz2mol(atoms, coordinates, charge=charge_read)
        with open(xyz_file) as f:
            xyz = f.read()

        _log.info(f"Generated {len(mols)} molecules.")

        if len(mols) > 0:
            mol = mols[0]
            return cls(mol=mol, xyz=xyz)

        else:
            raise IOError("Error creating molecule object from XYZ file: no \
molecules generated by xyz2mol.")

    @property
    def geometry(self):
        if self._xyz:
            return Geometry.from_xyz(self._xyz)
        return Geometry.from_mol(self.mol)

    def xyz(self, name: str = "unnamed"):
        if self._xyz:
            return self._xyz
        return self.geometry.to_xyz(title=name)

    @property
    def smiles(self):
        """Convert RDKit molecule to canonical Smiles string"""
        from rdkit.Chem import AllChem as Chem
        return Chem.MolToSmiles(Chem.RemoveHs(self.mol))

    @property
    def num_electrons(self):
        """Get the number of electrons for the molecule"""
        return num_electrons(self.mol)

    def all_atoms(self):
        return [atom.GetAtomicNum() for atom in self.mol.GetAtoms()]

    @property
    def atoms(self) -> List[int]:
        return sorted(set(self.all_atoms()))

    @property
    def atom_numbers(self) -> Dict[int, int]:
        """Get a dictionary of the atomic numbers of atoms in the molecule
        mapped to the amount of each in the molecule"""
        atoms = self.all_atoms()
        return {
            atomic_number:
                atoms.count(atomic_number) for atomic_number in set(atoms)
        }

    def basis(self, basis: str = "STO-3G"):
        # Obtain the basis set in nwchem format (as a string)
        atoms = self.atoms
        basis = bse.get_basis(basis, elements=atoms, fmt='nwchem')
        ob = "d*[spdf]"
        parsed_basis = re.findall(
            f"-> \\[(\\{ob},*\\{ob}*,*\\{ob}*)\\]\n(\\w+)",
            basis
        )
        assert len(atoms) == len(parsed_basis), \
            f"Cannot parse basis: number of atoms does not match result: \
{basis}"
        return dict(zip(atoms, parsed_basis))

    def num_orbitals(self, basis: str = "STO-3G") -> int:
        """Get total number of spatial orbitals for the molecule by
        using the basis states"""
        pattern = "(\\d)*s*,*(\\d)*p*,*(\\d)*d*,*(\\d)*f*"
        orbital_fillings = [1, 3, 5, 7]

        spin_orbitals = 0
        for atom, (basis, name) in self.basis(basis).items():
            num_atoms = self.atom_numbers[atom]
            s, p, d, f = re.findall(pattern, basis)[0]

            num_electrons_per_orbital = [
                filling * int(num) for num, filling in zip(
                    [s, p, d, f],
                    orbital_fillings
                ) if num
            ]
            num_electrons_per_atom = sum(num_electrons_per_orbital)
            spin_orbitals += num_atoms * num_electrons_per_atom

        return int(spin_orbitals)

    @classmethod
    def design(cls):
        from varname import varname, VarnameRetrievingError
        try:
            name = varname()
        except VarnameRetrievingError:
            name = "_"

        mol = cls(mol=None)
        mol.design_widget = JsmeWidget(parent_varname=name)
        mol._updated = True
        display(mol.design_widget)
        return mol

    def update_design(self, add_hs: bool = True, num_confs: int = 10):
        mol = self.design_widget.to_mol(add_hs=add_hs)
        self.mol = mol
        self.widget = JsmolWidget.from_mol(mol=mol, num_confs=num_confs)

    def create_input(
        self,
        molecule_name: str,
        file_name: str,
        solver: str,
        base_path: str = None,
        **parameters
    ):
        """Create input deck and save to file

        :param molecule_name: Name of the input deck
        :type molecule_name: str
        :param file_name: Output file name
        :type file_name: str
        :param solver: Solver to use
        :type solver: str
        :param base_path: Path to save output file to
        :type base_path: str
        :param parameters: Parameters for input deck
        :type parameters: dict
        """
        try:
            solver_spec = Solver[solver.lower()].value

        except Exception:
            names = [_s.name for _s in Solver]
            raise ValueError(
                f"Solver {solver} not found. Valid values: {names}"
            )

        else:
            if base_path is None:
                base_path = DEFAULT_BASE_PATH

            file_path = os.path.join(base_path, file_name)

            input_deck = solver_spec.module.create_input_deck(
                mol=self.mol,
                mol_name=molecule_name,
                geometry=self.geometry,
                **parameters
            )

            _log.info(
                f"Saving {solver_spec.name} input deck to file {file_path}"
            )
            with open(file_path, "w") as f:
                f.write(input_deck)

            return file_path

    def _ipython_display_(self):
        if self.design_widget and self.design_widget.was_updated:
            self.update_design()
            self.design_widget.reset_updated()

        return display(self.widget)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import warnings

from typing import Union, TYPE_CHECKING

from qdk.chemistry.geometry import (
    Geometry,
    format_geometry,
    format_geometry_from_mol
)

if TYPE_CHECKING:
    from rdkit.Chem.AllChem import Mol


def num_atoms_from_mol(mol: "Mol") -> int:
    """Calculate the number of atoms in the molecule

    :param mol: RDKit molecule object
    :type mol: Mol
    :return: Number of atoms or sum of atomic number of atoms in molecule
    :rtype: int
    """
    return len(mol.GetAtoms())


def num_electrons(mol: "Mol") -> int:
    """Calculate the number of electrons in the molecule

    :param mol: RDKit molecule object
    :type mol: Mol
    :return: Number of electrons or sum of atomic number of atoms in molecule
    :rtype: int
    """
    return sum([atom.GetAtomicNum() for atom in mol.GetAtoms()])


def formatted_geometry_str(
    mol: "Mol",
    geometry: Union[str, Geometry] = None,
    line_sep: str = "\n"
) -> str:
    """Format geometry for input deck

    :param mol: Molecule object
    :type mol: Mol, optional
    :param geometry: Geometry object or string, defaults to None
    :type geometry: Union[str, Geometry], optional
    :param line_sep: Line separator, defaults to "\n"
    :type line_sep: str
    :return: Formatted geometry string
    :rtype: str
    """
    if geometry is None:
        geometry = format_geometry_from_mol(mol=mol, line_sep=line_sep)
    else:
        warnings.warn(
            "Ignoring mol and using specified geometry string instead.")

        if isinstance(geometry, Geometry):
            geometry = format_geometry(geometry=geometry, line_sep=line_sep)

    return geometry
